﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing.Drawing2D;

namespace Function_Editor
{
    public partial class Function
    {
        public enum Types
        {
            Identity = 0,		// 0
            Constant = 1,		// 0
            Transition = 2,		// 0
            Periodic = 3,		// 0
            Linear = 4,			// 2
            LinearKey = 5,		// 4
            //MultiLinearKey = 6,	// 16
            Spline = 7,			// 4
            //MultiSpline = 8,	// 16
            Exponent = 9,		// 0
            Spline2 = 10,		// 4
        }

        public static readonly int[] TypeSizeTable =
        {
			0,
			1,
			2,
			4,
			6,
			20,
			32,
			12,
			4,
			3,
			12,
		};

        public enum FlagOptions
        {
            Range = 1
        }

        public enum OutputTypes
        {
            Scalar,
            Constant,
            Color2,
            Color3,
            Color4,
        }

        public enum TransitionExponents
        {
            Linear,
            Early,
            VeryEarly,
            Late,
            VeryLate,
            Cosine,
            One,
            Zero,
        }

        public enum PeriodicExponents
        {
            One,
            Zero,
            Cosine,
            CosineVariablePeriod,
            DiagonalWave,
            DiagonalWaveVariablePeriod,
            Slide,
            SlideVariablePeriod,
            Noise,
            Jitter,
            Wander,
            Spark,
        }

        public static byte GetPathPointType(Types t)
        {
            switch (t)
            {
                case Types.Identity: 
                case Types.Constant:
                case Types.Linear:
                case Types.LinearKey:
              //case Types.MultiLinearKey:
                case Types.Exponent:
                case Types.Transition:
                case Types.Periodic:
                    return (byte)PathPointType.Line;
                case Types.Spline:
              //case Types.MultiSpline:
                    return (byte)PathPointType.Bezier;
                case Types.Spline2:
                    return (byte)PathPointType.Bezier3;
            }
            throw new Exception("Unknown Type: " + t.ToString());
        }

        public static int GetNumberOfPoints(Types t)
        {
            switch (t)
            {
                case Types.Identity: return 0;
                case Types.Constant: return 0;
                case Types.Transition: return 0;
                case Types.Periodic: return 0;
                case Types.Linear: return 2;
                case Types.LinearKey: return 4;
                //case Types.MultiLinearKey: return 16;
                case Types.Spline: return 4;
                //case Types.MultiSpline: return 8;
                case Types.Exponent: return 0;
                case Types.Spline2: return 4;
            }
            throw new Exception("Unknown Type: " + t.ToString());
        }

        public static int GetDataSize(Types t)
        {
            return (TypeSizeTable[(int)t] * 8) + 20;
        }
    }
}
