﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Function_Editor
{
    public partial class FunctionEditor : UserControl
    {
        public Function Function { get { return function; } }
        private Function function = new Function();

        public FunctionEditor()
        {
            InitializeComponent();

            cboType.DataSource = Enum.GetValues(typeof(Function.Types));
            cboOutputType.DataSource = Enum.GetValues(typeof(Function.OutputTypes));

            colorPickerUpper.Tag = 3;

            colorPickerUpperMid.Tag = 2;
            colorPickerMid.Tag = 1;
            colorPickerLowerMid.Tag = 1;

            colorPickerLower.Tag = 0;
        }

        private void FunctionEditor_Load(object sender, EventArgs e)
        { RefreshFunction(); }

        bool isRefreshing = false;
        public void RefreshFunction()
        {
            isRefreshing = true;
            switch (function.Type)
            {
                case Function.Types.Identity:
                case Function.Types.Constant:
                    graphEditor1.GraphPoints1 = new PointF[] { new PointF(0.0f, 0.0f), new PointF(1.0f, 1.0f) };
                    graphEditor1.GraphPoints2 = new PointF[] { new PointF(0.0f, 0.0f), new PointF(1.0f, 1.0f) };
                    break;
                case Function.Types.Exponent:
                    graphEditor1.GraphPoints1 = GetGraphPoints(function.Exponent1);
                    graphEditor1.GraphPoints2 = GetGraphPoints(function.Exponent2);
                    break;
                case Function.Types.Transition:
                    graphEditor1.GraphPoints1 = GetGraphPoints((Function.TransitionExponents)Function.ExponentType1, 0.0f, 1.0f);
                    graphEditor1.GraphPoints2 = GetGraphPoints((Function.TransitionExponents)Function.ExponentType1, 0.0f, 1.0f);
                    break;
                case Function.Types.Periodic:
                    graphEditor1.GraphPoints1 = GetGraphPoints((Function.PeriodicExponents)Function.ExponentType1, 0.0f, 1.0f);
                    graphEditor1.GraphPoints2 = GetGraphPoints((Function.PeriodicExponents)Function.ExponentType1, 0.0f, 1.0f);
                    break;
                case Function.Types.Linear:
                case Function.Types.LinearKey:
                //case Types.MultiLinearKey:
                case Function.Types.Spline:
                case Function.Types.Spline2:
                    graphEditor1.GraphPoints1 = function.Points1;
                    graphEditor1.GraphPoints2 = function.Points2;
                    break;
            }

            if (function.Points1 != null)
                graphEditor1.GraphPointTypes1 = GetPointTypeArray(Function.GetPathPointType(function.Type), graphEditor1.GraphPoints1.Length);

            if (function.Points2 != null)
                graphEditor1.GraphPointTypes2 = GetPointTypeArray(Function.GetPathPointType(function.Type), graphEditor1.GraphPoints2.Length);

            cboType.SelectedItem = function.Type;
            cboOutputType.SelectedItem = function.OutputType;

            ShowControls((Function.Types)cboType.SelectedItem, (Function.OutputTypes)cboOutputType.SelectedItem, function.IsRange);

            cbRange.Checked = function.IsRange;
            graphEditor1.Range = function.IsRange;
            
            numExponent.Value = (decimal)function.Exponent1;
            numExponent2.Value = (decimal)function.Exponent2;

            numFunctionMin.Value = (decimal)function.FunctionBounds1.X;
            numFunctionMin2.Value = (decimal)function.FunctionBounds2.X;

            numFunctionMax.Value = (decimal)function.FunctionBounds1.Y;
            numFunctionMax2.Value = (decimal)function.FunctionBounds2.Y;

            numFrequency.Value = (decimal)function.Frequency1;
            numFrequency2.Value = (decimal)function.Frequency2;

            numPhase.Value = (decimal)function.Phase1;
            numPhase2.Value = (decimal)function.Phase2;

            switch (function.Type)
            {
                case Function.Types.Periodic:
                {
                    cboExponentType.DataSource = Enum.GetValues(typeof(Function.PeriodicExponents));
                    cboExponentType.SelectedIndex = function.ExponentType1;

                    cboExponentType2.DataSource = Enum.GetValues(typeof(Function.PeriodicExponents));
                    cboExponentType2.SelectedIndex = function.ExponentType2;
                    break;
                }
                case Function.Types.Transition:
                {
                    cboExponentType.DataSource = Enum.GetValues(typeof(Function.TransitionExponents));
                    cboExponentType.SelectedIndex = function.ExponentType1;

                    cboExponentType2.DataSource = Enum.GetValues(typeof(Function.TransitionExponents));
                    cboExponentType2.SelectedIndex = function.ExponentType2;
                    break;
                }
            }

            if (function.OutputType == Function.OutputTypes.Scalar)
            {
                numMinY.Value = (decimal)function.YBounds.X;
                numMaxY.Value = (decimal)function.YBounds.Y;
            }

            RefreshGraph();
            isRefreshing = false;
        }

        void RefreshGradient()
        {
            gradientViewer1.GradientBitmap = graphEditor1.OutputGradientBitmap;
            gradientViewer1.Invalidate();
        }

        void RefreshGraph()
        {
            graphEditor1.GraphColors = GetUsedColors();
            graphEditor1.Invalidate();
        }

        Color[] GetUsedColors()
        {
            switch (function.OutputType)
            {
                case Function.OutputTypes.Scalar:
                    return new Color[] { Color.Black, Color.White };
                case Function.OutputTypes.Constant:
                    return new Color[] { function.Colors[0] };
                case Function.OutputTypes.Color2:
                    return new Color[] { function.Colors[0], function.Colors[3] };
                case Function.OutputTypes.Color3:
                    return new Color[] { function.Colors[0], function.Colors[1], function.Colors[3] };
                case Function.OutputTypes.Color4:
                    return new Color[] { function.Colors[0], function.Colors[1], function.Colors[2], function.Colors[3] };
            }
            return new Color[0];
        }

        #region GetGraphPoints
        private PointF[] GetGraphPoints(Function.PeriodicExponents periodicExponent, float min, float max)
        {
            float scale = max - min;
            PointF[] points = new PointF[graphEditor1.Width / 10];
            for (int i = 0; i < points.Length; i++)
            {
                float x = ((float)i / (float)(points.Length - 1));

                switch (periodicExponent)
                {
                    case Function.PeriodicExponents.One:
                        points[i] = new PointF(x, max);
                        break;
                    case Function.PeriodicExponents.Zero:
                        points[i] = new PointF(x, min);
                        break;
                    case Function.PeriodicExponents.Cosine:
                        points[i] = new PointF(x, 0.5f + min + ((float)Math.Cos((double)x * 3.14f * 2) * scale * 0.5f));
                        break;
                    case Function.PeriodicExponents.CosineVariablePeriod:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.DiagonalWave:
                        points[i] = new PointF(x, min + (x * scale));
                        break;
                    case Function.PeriodicExponents.DiagonalWaveVariablePeriod:
                        points[i] = new PointF(x, 1.0f);
                        break;
                    case Function.PeriodicExponents.Slide:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.SlideVariablePeriod:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.Noise:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.Jitter:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.Wander:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.PeriodicExponents.Spark:
                        points[i] = new PointF(x, 0.0f);
                        break;
                }
            }
            return points;
        }

        private PointF[] GetGraphPoints(Function.TransitionExponents transitionExponent, float min, float max)
        {
            float scale = max - min;
            PointF[] points = new PointF[graphEditor1.Width / 10];
            for (int i = 0; i < points.Length; i++)
            {
                float x = ((float)i / (float)(points.Length - 1));

                switch (transitionExponent)
                { 
                    case Function.TransitionExponents.Cosine:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.TransitionExponents.Early:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.TransitionExponents.Late:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.TransitionExponents.Linear:
                        points[i] = new PointF(x, min + (x * scale));
                        break;
                    case Function.TransitionExponents.One:
                        points[i] = new PointF(x, max);
                        break;
                    case Function.TransitionExponents.VeryEarly:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.TransitionExponents.VeryLate:
                        points[i] = new PointF(x, 0.0f);
                        break;
                    case Function.TransitionExponents.Zero:
                        points[i] = new PointF(x, min);
                        break;
                }
            }
            return points;
        }

        private PointF[] GetGraphPoints(float exponent)
        {
            PointF[] points = new PointF[graphEditor1.Width / 10];
            for (int i = 0; i < points.Length; i++)
            {
                float x = ((float)i / (float)(points.Length - 1));
                points[i] = new PointF(x, (float)Math.Pow((double)x, (double)exponent));
            }
            return points;
        }

        byte[] GetPointTypeArray(byte type, int size)
        {
            byte[] types = new byte[size];
            for (int i = 0; i < size; i++)
                types[i] = type;
            return types;
        }
        #endregion
       
        #region Show/Hide Controls
        void ShowControls(Function.Types type, Function.OutputTypes outputType, bool range)
        {
            switch (type)
            {
                case Function.Types.Identity:
                case Function.Types.Constant:
                    SetVisibleExponent(false, range);
                    SetVisibleExponentType(false, range);
                    SetVisibleFunctionRange(false, range);
                    SetVisibleFrequencyAndPhase(false, range);
                    graphEditor1.UserCanEdit = false;
                    break;
                case Function.Types.Exponent:
                    SetVisibleExponentType(false, range);
                    SetVisibleExponent(true, range);
                    SetVisibleFunctionRange(true, range);
                    SetVisibleFrequencyAndPhase(false, range);
                    graphEditor1.UserCanEdit = false;
                    break;
                case Function.Types.Linear:
                case Function.Types.LinearKey:
                //case Function.Types.MultiLinearKey:
                case Function.Types.Spline:
                case Function.Types.Spline2:
                //case Function.Types.MultiSpline:
                    SetVisibleExponent(false, range);
                    SetVisibleExponentType(false, range);
                    SetVisibleFunctionRange(false, range);
                    SetVisibleFrequencyAndPhase(false, range);
                    graphEditor1.UserCanEdit = true;
                    break;
                case Function.Types.Periodic:
                    SetVisibleExponent(false, range);
                    SetVisibleExponentType(true, range);
                    SetVisibleFunctionRange(true, range);
                    SetVisibleFrequencyAndPhase(true, range);
                    graphEditor1.UserCanEdit = false;
                    break;
                case Function.Types.Transition:
                    SetVisibleExponent(false, range);
                    SetVisibleExponentType(true, range);
                    SetVisibleFunctionRange(true, range);
                    SetVisibleFrequencyAndPhase(false, range);
                    graphEditor1.UserCanEdit = false;
                    break;
            }

            switch (outputType)
            {
                case Function.OutputTypes.Scalar:
                    SetVisibleYRange(true, range);
                    SetVisibleColorLower(false);
                    SetVisibleColorLowerMid(false);
                    SetVisibleColorMid(false);
                    SetVisibleColorUpperMid(false);
                    SetVisibleColorUpper(false);
                    break;
                case Function.OutputTypes.Constant:
                    SetVisibleYRange(false, range);
                    SetVisibleColorLower(true);
                    SetVisibleColorLowerMid(false);
                    SetVisibleColorMid(false);
                    SetVisibleColorUpperMid(false);
                    SetVisibleColorUpper(false);
                    break;
                case Function.OutputTypes.Color2:
                    SetVisibleYRange(false, range);
                    SetVisibleColorLower(true);
                    SetVisibleColorLowerMid(false);
                    SetVisibleColorMid(false);
                    SetVisibleColorUpperMid(false);
                    SetVisibleColorUpper(true);
                    lblColorUpper.Text = "Color 2";
                    break;
                case Function.OutputTypes.Color3:
                    SetVisibleYRange(false, range);
                    SetVisibleColorLower(true);
                    SetVisibleColorLowerMid(false);
                    SetVisibleColorMid(true);
                    SetVisibleColorUpperMid(false);
                    SetVisibleColorUpper(true);
                    lblColorUpper.Text = "Color 3";
                    break;
                case Function.OutputTypes.Color4:
                    SetVisibleYRange(false, range);
                    SetVisibleColorLower(true);
                    SetVisibleColorLowerMid(true);
                    SetVisibleColorMid(false);
                    SetVisibleColorUpperMid(true);
                    SetVisibleColorUpper(true);
                    lblColorUpper.Text = "Color 4";
                    break;
            }

            colorPickerLower.SelectedColor = function.Colors[(int)colorPickerLower.Tag];

            colorPickerMid.SelectedColor = function.Colors[(int)colorPickerMid.Tag];
            colorPickerLowerMid.SelectedColor = function.Colors[(int)colorPickerLowerMid.Tag];

            colorPickerUpperMid.SelectedColor = function.Colors[(int)colorPickerUpperMid.Tag];

            colorPickerUpper.SelectedColor = function.Colors[(int)colorPickerUpper.Tag]; 
        }

        private void SetVisibleColorUpper(bool visible)
        {
            lblColorUpper.Visible = visible;
            colorPickerUpper.Visible = visible;
        }

        private void SetVisibleColorUpperMid(bool visible)
        {
            lblColorUpperMid.Visible = visible;
            colorPickerUpperMid.Visible = visible;
        }

        private void SetVisibleColorMid(bool visible)
        {
            lblColorMid.Visible = visible;
            colorPickerMid.Visible = visible;
        }

        private void SetVisibleColorLowerMid(bool visible)
        {
            lblColorLowerMid.Visible = visible;
            colorPickerLowerMid.Visible = visible;
        }

        private void SetVisibleColorLower(bool visible)
        {
            lblColorLower.Visible = visible;
            colorPickerLower.Visible = visible;
        }

        private void SetVisibleFunctionRange(bool visible, bool range)
        {
            numFunctionMin.Visible = visible;
            numFunctionMax.Visible = visible;
            lblFunctionMin.Visible = visible;
            lblFunctionMax.Visible = visible;

            numFunctionMin2.Visible = visible && range;
            numFunctionMax2.Visible = visible && range;
            lblFunctionMin2.Visible = visible && range;
            lblFunctionMax2.Visible = visible && range;
        }

        private void SetVisibleYRange(bool visible, bool range)
        {
            numMinY.Visible = visible;
            numMaxY.Visible = visible;
            lblYMin.Visible = visible;
            lblYMax.Visible = visible;
        }

        private void SetVisibleExponent(bool visible, bool range)
        {
            numExponent.Visible = visible;
            lblExponent.Visible = visible;

            numExponent2.Visible = visible && range;
            lblExponent2.Visible = visible && range;
        }

        private void SetVisibleExponentType(bool visible, bool range)
        {
            cboExponentType.Visible = visible;
            lblExponent.Visible = visible;

            cboExponentType2.Visible = visible && range;
            lblExponent2.Visible = visible && range;
        }

        private void SetVisibleFrequencyAndPhase(bool visible, bool range)
        {
            numFrequency.Visible = visible;
            numPhase.Visible = visible;
            lblFrequency.Visible = visible;
            lblPhase.Visible = visible;

            numFrequency2.Visible = visible && range;
            numPhase2.Visible = visible && range;
            lblFrequency2.Visible = visible && range;
            lblPhase2.Visible = visible && range;
        }
#endregion

        private void graphEditor1_MouseMove(object sender, MouseEventArgs e)
        {
            PointF point = graphEditor1.Constrain(graphEditor1.ToGraphCoords(e.Location));
            lblCoords.Text = string.Format("X: {0}\nY: {1}", Math.Round(point.X, 2), Math.Round(point.Y, 2));

            if(e.Button == MouseButtons.Left)
                RefreshGradient();
        }

        private void FunctionEditor_Resize(object sender, EventArgs e)
        {
            int offset = (graphEditor1.Height - (graphEditor1.Border.Height * 2)) / 3;

            colorPickerUpper.Top = graphEditor1.Top - (colorPickerUpper.Size.Height / 2) + graphEditor1.Border.Height;
            lblColorUpper.Top = graphEditor1.Top - (lblColorUpper.Size.Height / 2) + graphEditor1.Border.Height;

            colorPickerUpperMid.Top = graphEditor1.Top + offset - (colorPickerUpperMid.Size.Height / 2) + graphEditor1.Border.Height;
            lblColorUpperMid.Top = graphEditor1.Top + offset - (lblColorUpperMid.Size.Height / 2) + graphEditor1.Border.Height;

            colorPickerMid.Top = graphEditor1.Top + (graphEditor1.Height / 2) - (colorPickerMid.Size.Height / 2);
            lblColorMid.Top = graphEditor1.Top + (graphEditor1.Height / 2) - (lblColorMid.Size.Height / 2);

            colorPickerLowerMid.Top = graphEditor1.Top + (offset * 2) - (colorPickerLowerMid.Size.Height / 2) + graphEditor1.Border.Height;
            lblColorLowerMid.Top = graphEditor1.Top + (offset * 2) - (lblColorLowerMid.Size.Height / 2) + graphEditor1.Border.Height;

            colorPickerLower.Top = graphEditor1.Bottom - (colorPickerLower.Size.Height / 2) - graphEditor1.Border.Height;
            lblColorLower.Top = graphEditor1.Bottom - (lblColorLower.Size.Height / 2) - graphEditor1.Border.Height;

            cbRange.Left = graphEditor1.Left + (graphEditor1.Width / 2) - (cbRange.Width / 2);

            RefreshGradient();
        }

        private void graphEditor1_Paint(object sender, PaintEventArgs e)
        { RefreshGradient(); }

        private void cboType_SelectedValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Type = (Function.Types)cboType.SelectedValue;
            ShowControls((Function.Types)cboType.SelectedItem, (Function.OutputTypes)cboOutputType.SelectedItem, cbRange.Checked);
            RefreshFunction();
        }

        private void cboOutputType_SelectedValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.OutputType = (Function.OutputTypes)cboOutputType.SelectedValue;
            ShowControls((Function.Types)cboType.SelectedItem, (Function.OutputTypes)cboOutputType.SelectedItem, cbRange.Checked);
            RefreshFunction();
        }

        private void colorPicker_SelectedColorChanged(object sender, EventArgs e)
        {
            ColorPicker picker = (ColorPicker)sender;
            function.Colors[(int)picker.Tag] = picker.SelectedColor;
            RefreshGraph();
            RefreshGradient();
        }

        private void cboExponentType_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.ExponentType1 = cboExponentType.SelectedIndex;
            RefreshFunction();
        }

        private void cboExponentType2_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.ExponentType2 = cboExponentType2.SelectedIndex;
            RefreshFunction();
        }

        private void cbRange_CheckedChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.IsRange = cbRange.Checked;
            RefreshFunction();
        }


        #region YBounds
        private void numMinY_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.YBounds = new PointF((float)numMinY.Value, function.YBounds.Y);
            RefreshFunction();
        }

        private void numMaxY_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.YBounds = new PointF(function.YBounds.X, (float)numMaxY.Value);
            RefreshFunction();
        }
        #endregion

        #region Function 1
        private void numFrequency_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Frequency1 = (float)numFrequency.Value;
            RefreshFunction();
        }

        private void numFunctionMin_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.FunctionBounds1 = new PointF((float)numFunctionMin.Value, function.FunctionBounds1.Y);
            RefreshFunction();
        }

        private void numFunctionMax_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.FunctionBounds1 = new PointF(function.FunctionBounds1.X, (float)numFunctionMax.Value);
            RefreshFunction();
        }


        private void numPhase_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Phase1 = (float)numPhase.Value;
            RefreshFunction();
        }

        private void numExponent_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Exponent1 = (float)numExponent.Value;
            RefreshFunction();
        }
#endregion

        #region Function 2
        private void numFrequency2_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Frequency2 = (float)numFrequency2.Value;
            RefreshFunction();
        }

        private void numFunctionMin2_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.FunctionBounds2 = new PointF((float)numFunctionMin2.Value, function.FunctionBounds2.Y);
            RefreshFunction();
        }

        private void numFunctionMax2_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.FunctionBounds2 = new PointF(function.FunctionBounds2.X, (float)numFunctionMax2.Value);
            RefreshFunction();
        }

        private void numPhase2_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Phase2 = (float)numPhase2.Value;
            RefreshFunction();
        }

        private void numExponent2_ValueChanged(object sender, EventArgs e)
        {
            if (isRefreshing) return;
            function.Exponent2 = (float)numExponent2.Value;
            RefreshFunction();
        }
        #endregion
    }
}
