﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

namespace Function_Editor
{
    public partial class GraphEditor : UserControl
    {
        Pen ForeColorPen;
        Pen Graph1Pen;
        Pen Graph2Pen;
        SolidBrush HitBoxBrush1;
        SolidBrush HitBoxBrush2;

        public int HorziontalGridLineCount { get; set; }
        public int VerticalGridLineCount { get; set; }

        public Color[] GraphColors { get; set; }

        public PointF[] GraphPoints1 { get; set; }
        public byte[] GraphPointTypes1 { get; set; }
        
        public PointF[] GraphPoints2 { get; set; }
        public byte[] GraphPointTypes2 { get; set; }

        public bool UserCanEdit { get; set; }

        public bool Range { get; set; }

        public Size Border { get { return border; } }
        private Size border = new Size(10, 10);

        Bitmap graph1Bitmap;
        Bitmap graph2Bitmap;
        Bitmap gradientBitmap;

        public Bitmap OutputGradientBitmap { get { return outputGradientBitmap; } }
        private Bitmap outputGradientBitmap;

        public GraphEditor()
        {
            InitializeComponent();

            ForeColorPen = new Pen(ForeColor);
            Graph1Pen = new Pen(Color.Green);
            Graph2Pen = new Pen(Color.Red);
            HitBoxBrush1 = new SolidBrush(Color.Green);
            HitBoxBrush2 = new SolidBrush(Color.Red);

            GraphColors = new Color[2] { Color.White, Color.Black };
            
            GraphPoints1 = new PointF[0];
            GraphPointTypes1 = new byte[0];

            GraphPoints2 = new PointF[0];
            GraphPointTypes2 = new byte[0];

            HorziontalGridLineCount = 4;
            VerticalGridLineCount = 9;

            graph1Bitmap = new Bitmap(Width, Height);
            graph2Bitmap = new Bitmap(Width, Height);
            gradientBitmap = new Bitmap(Width, Height);
            outputGradientBitmap = new Bitmap(Width, 2);
            GenerateGradientBitmap();
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            GenerateGradientBitmap();

            DrawBackground(e.Graphics);
            DrawGridLines(e.Graphics);

            if (GraphPoints1 != null && GraphPointTypes1 != null)
            {
                DrawGraph(e.Graphics, Graph1Pen, GraphPoints1, GraphPointTypes1);
                DrawGraphPoints(e.Graphics, HitBoxBrush1, GraphPoints1);
            }

            if (Range && GraphPoints2 != null && GraphPointTypes2 != null)
            {
                DrawGraph(e.Graphics, Graph2Pen, GraphPoints2, GraphPointTypes2);
                DrawGraphPoints(e.Graphics, HitBoxBrush2, GraphPoints2);
            }
            
            base.OnPaint(e);
        }

        private void GenerateGradientBitmap()
        {
            Graphics graph1Graphics = Graphics.FromImage(graph1Bitmap);
            graph1Graphics.Clear(Color.FromArgb(0));
            DrawGraph(graph1Graphics, Graph1Pen, GraphPoints1, GraphPointTypes1);

            Graphics graph2Graphics = Graphics.FromImage(graph2Bitmap);
            graph2Graphics.Clear(Color.FromArgb(0));
            if(Range)
                DrawGraph(graph2Graphics, Graph2Pen, GraphPoints2, GraphPointTypes2);

            Graphics gradientGraphics = Graphics.FromImage(gradientBitmap);
            gradientGraphics.Clear(Color.FromArgb(0));
            DrawBackground(gradientGraphics);

            unsafe
            {
                int graph1Color = Graph1Pen.Color.ToArgb();
                int graph2Color = Graph2Pen.Color.ToArgb();
                
                BitmapData graph1BD = graph1Bitmap.LockBits(new Rectangle(0, 0, Width, Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                BitmapData graph2BD = graph2Bitmap.LockBits(new Rectangle(0, 0, Width, Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                BitmapData gradientBD = gradientBitmap.LockBits(new Rectangle(0, 0, Width, Height), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                BitmapData outputBD = outputGradientBitmap.LockBits(new Rectangle(0, 0, Width, 1), ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                
                int* graph1Start = (int*)graph1BD.Scan0.ToPointer();
                int* graph2Start = (int*)graph2BD.Scan0.ToPointer();
                int* gradientStart = (int*)gradientBD.Scan0.ToPointer();
                int* outputStart = (int*)outputBD.Scan0.ToPointer();

                for (int x = 0; x < Width; x++)
                {
                    bool found1 = false;
                    bool found2 = false;
                    int* outputPtr = outputStart + x;
                    for (int y = 0; y < Height; y++)
                    {
                        int offset = (y * Width) + x;
                        int* cPtr = graph1Start + offset;
                        if (*cPtr == graph1Color)
                        {
                            *outputPtr = *(gradientStart + offset);
                            found1 = true;
                            if (!Range)
                            {
                                *(outputPtr + Width) = *(gradientStart + offset);
                                found2 = true;
                            }
                        }

                        cPtr = graph2Start + offset;
                        if (Range && *cPtr == graph2Color)
                        {
                            *(outputPtr + Width) = *(gradientStart + offset);
                            found2 = true;
                        }

                        if (found1 && (!Range || found2)) break;
                    }
                    if (!found1)
                        *outputPtr = BackColor.ToArgb();
                    if (!found2)
                        *(outputPtr + Width) = BackColor.ToArgb();
                }
                graph1Bitmap.UnlockBits(graph1BD);
                graph2Bitmap.UnlockBits(graph2BD);
                gradientBitmap.UnlockBits(gradientBD);
                outputGradientBitmap.UnlockBits(outputBD);
            }
        }

        private void DrawBackground(Graphics g)
        {
            if (GraphColors.Length == 0)
                GraphColors = new Color[] { Color.Black };

            RectangleF rect = new RectangleF(border.Width, border.Height - 1, Width - (border.Width * 2), Height - (border.Height * 2) + 2);
            if (rect.Width == 0 || rect.Height == 0) return;

            if (GraphColors.Length == 1)
                GraphColors = new Color[] { GraphColors[0], GraphColors[0] };

            ColorBlend blend = new ColorBlend(GraphColors.Length);
            blend.Colors = GraphColors;
            for (int i = 0; i < GraphColors.Length; i++)
                blend.Positions[i] = (float)i / ((float)GraphColors.Length - 1);
         
            LinearGradientBrush brush = new LinearGradientBrush(rect, GraphColors[0], GraphColors[1], 270);
            brush.InterpolationColors = blend;

            g.FillRectangle(brush, rect);
        }

        private void DrawGraph(Graphics g, Pen pen, PointF[] graphPoints, byte[] graphPointTypes)
        {
            if (graphPoints == null || graphPointTypes == null || graphPoints.Length == 0 || graphPointTypes.Length != graphPoints.Length) return;
            g.DrawPath(pen, GetGraphicsPathScreenCoords(graphPoints, graphPointTypes));
        }

        public GraphicsPath GetGraphicsPathScreenCoords(PointF[] graphPoints, byte[] graphPointTypes)
        {
            PointF[] screenPoints = new PointF[graphPoints.Length];
            for (int i = 0; i < screenPoints.Length; i++)
                screenPoints[i] = ToScreenCoords(graphPoints[i]);
            return new GraphicsPath(screenPoints, graphPointTypes);
        }

        private void DrawGraphPoints(Graphics g, Brush brush, PointF[] graphPoints)
        {
            if (!UserCanEdit) return;
            foreach (PointF p in graphPoints)
            {
                RectangleF rect = GetHitBox(p);
                g.FillRectangle(brush, rect);
                g.DrawRectangle(Pens.Black, rect.X, rect.Y, rect.Width, rect.Height);
            }
        }

        private void DrawGridLines(Graphics g)
        {
            g.DrawRectangle(ForeColorPen, border.Width, border.Height, GraphWidth(), GraphHeight());

            float division = GraphHeight() / (float)(HorziontalGridLineCount + 1);
            for (int i = 1; i < (HorziontalGridLineCount + 1); i++)
                g.DrawLine(ForeColorPen,
                    new PointF(0.0f, division * (float)i + border.Height),
                    new PointF(Width, division * (float)i + border.Height));

            division = GraphWidth() / (float)(VerticalGridLineCount + 1);
            for (int i = 1; i < (VerticalGridLineCount + 1); i++)
                g.DrawLine(ForeColorPen,
                    new PointF(division * (float)i + border.Width, 0.0f),
                    new PointF(division * (float)i + border.Width, Height));
        }

        int selectedPoint = -1;
        private void GraphEditor_MouseDown(object sender, MouseEventArgs e)
        {
            if (!UserCanEdit) return;
            selectedPoint = PointAt(new PointF(e.X, e.Y));
        }

        private void GraphEditor_MouseMove(object sender, MouseEventArgs e)
        {
            if (!UserCanEdit) return;

            if (PointAt(new PointF(e.X, e.Y)) == -1)
                Cursor = Cursors.Arrow;
            else
                Cursor = Cursors.Cross;

            if (selectedPoint == -1) return;
            PointF newPos = Constrain(ToGraphCoords(e.Location));
            if (selectedPoint == 0 || selectedPoint == GraphPoints1.Length)
                newPos.X = 0;
            else if (selectedPoint == GraphPoints1.Length - 1 || selectedPoint == GraphPoints1.Length + GraphPoints2.Length - 1)
                newPos.X = 1.0f;

            if (selectedPoint < GraphPoints1.Length)
                GraphPoints1[selectedPoint] = newPos;
            else
                GraphPoints2[selectedPoint - GraphPoints1.Length] = newPos;
            this.Invalidate();
        }

        public PointF Constrain(PointF point)
        {
            if (point.X < 0.0f) point.X = 0.0f;
            else if (point.X > 1.0f) point.X = 1.0f;
            if (point.Y < 0.0f) point.Y = 0.0f;
            else if (point.Y > 1.0f) point.Y = 1.0f;
            return point;
        }

        public PointF ToScreenCoords(PointF point)
        { 
            return new PointF(
                border.Width + point.X * GraphWidth(),
                border.Height + (1.0f - point.Y) * GraphHeight()
                );
        }

        public PointF ToGraphCoords(Point point)
        { 
            return new PointF(
                (float)(point.X - border.Width) / GraphWidth(),
                1.0f - (float)(point.Y - border.Height) / GraphHeight()
                ); 
        }

        private float GraphHeight()
        { return (float)(Height - (border.Height * 2)); }

        private float GraphWidth()
        { return (float)(Width - (border.Width * 2)); }

        private void GraphEditor_MouseUp(object sender, MouseEventArgs e)
        { selectedPoint = -1; }

        private RectangleF GetHitBox(PointF p)
        {
            p = ToScreenCoords(p);
            return new RectangleF(p.X - 5.0f, p.Y - 5.0f, 10.0f, 10.0f);
        }

        private int PointAt(PointF screenPoint)
        {
            for (int i = 0; i < GraphPoints1.Length; i++)
                if (GetHitBox(GraphPoints1[i]).Contains(screenPoint))
                    return i;
            if (Range)
            {
                for (int i = 0; i < GraphPoints2.Length; i++)
                    if (GetHitBox(GraphPoints2[i]).Contains(screenPoint))
                        return GraphPoints1.Length + i;
            }
            return -1;
        }

        private void GraphEditor_Resize(object sender, EventArgs e)
        {
            if (Width == 0 || Height == 0) return;
            
            graph1Bitmap.Dispose();
            graph2Bitmap.Dispose();
            gradientBitmap.Dispose();
            outputGradientBitmap.Dispose();

            graph1Bitmap = new Bitmap(Width, Height);
            graph2Bitmap = new Bitmap(Width, Height);
            gradientBitmap = new Bitmap(Width, Height);
            outputGradientBitmap = new Bitmap(Width, 2);

            GenerateGradientBitmap();

            this.Invalidate(); 
        }
    }
}
