﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using H2.Streams;
using System.IO;
using Function_Editor;

namespace Squiggle
{
    public partial class Main : Form
    {
        MapStream map;
        TagBlock selectedTagBlock;

        public Main()
        {
            InitializeComponent();

            cboType.SelectedItem = "prt3";
        }

        private void cmdOpenMap_Click(object sender, EventArgs e)
        {
            if (OFD.ShowDialog() != DialogResult.OK) return;
            map = new MapStream(OFD.FileName, null);
            lstTags.Sorted = true;
            lstTags.Items.Clear();
            lstTags.Items.AddRange(map.Tags.GetStringArrayOfClass((string)cboType.SelectedItem));
            map.Close();
        }

        private void cboType_SelectedValueChanged(object sender, EventArgs e)
        {
            cboChunks.Enabled = (string)cboType.SelectedItem != "prt3";

            if (map == null) return;
            map.ReOpen();
            lstTags.Items.Clear();
            lstTags.Items.AddRange(map.Tags.GetStringArrayOfClass((string)cboType.SelectedItem));
            map.Close();
        }

        private void lstTags_SelectedValueChanged(object sender, EventArgs e)
        {
            int selectedIndex = cboChunks.SelectedIndex;

            map.ReOpen();
            selectedTagBlock = TagBlock.Read(map, (string)cboType.SelectedItem, (string)lstTags.SelectedItem);
            map.Close();

            if (selectedTagBlock == null) return;

            cboChunks.Items.Clear();
            foreach(TagBlock.FunctionInfo f in selectedTagBlock.Functions)
                cboChunks.Items.Add(f);

            cboChunks.Enabled = cboChunks.Items.Count > 0;
            functionEditor1.Enabled = cboChunks.Enabled;
            if (cboChunks.Enabled)
            {
                if (selectedIndex >= cboChunks.Items.Count || selectedIndex < 0)
                    cboChunks.SelectedIndex = 0;
                else
                    cboChunks.SelectedIndex = selectedIndex;
            }
        }

        private void cboChunk_SelectedIndexChanged(object sender, EventArgs e)
        {
            LoadFunction(cboChunks.SelectedIndex);
        }

        private void cmdSave_Click(object sender, EventArgs e)
        {
            SaveFunction(cboChunks.SelectedIndex);
        }

        void LoadFunction(int index)
        {
            if (selectedTagBlock.Functions[index].DataLength == 0)
            {
                functionEditor1.Enabled = false;
                return;
            }
            
            functionEditor1.Enabled = true;
            map.ReOpen();
            selectedTagBlock.Functions[index].Load(map, functionEditor1.Function);
            map.Close();
            functionEditor1.RefreshFunction();
        }

        void SaveFunction(int index)
        {
            map.ReOpen();
            selectedTagBlock.Functions[index].Save(map, functionEditor1.Function);
            map.Close();

            LoadFunction(index);
        }
    }
}
